"""The Config sub-package sets up the environment for adalib. It also fetches all the
configuration values needed in the library.

The first client call to import this module will initialize it, after which it
will be a singleton instantiation for the duration of the process.
"""

from .config import Configuration, get_config

__all__ = ["Configuration", "get_config"]

__title__ = "adalib-auth Config"



"""Keycloak authentication module for adalib-auth."""

from .. import config, jupyterhub
from .keycloak import (
    app_authentication,
    authenticate_with_stored_credentials,
    external_authentication,
    get_client,
    get_client_token,
    get_token_from_exchange,
    jupyterhub_authentication,
    update_tokens_in_config,
)

__all__ = [
    "app_authentication",
    "authenticate_with_stored_credentials",
    "external_authentication",
    "get_client",
    "get_client_token",
    "get_token_from_exchange",
    "jupyterhub_authentication",
    "update_tokens_in_config",
    "config",
    "jupyterhub",
]

__title__ = "adalib-auth Keycloak"



"""The adalib-auth package handles the authentication workflow with the AdaLab platform.
"""

import importlib.metadata

from . import config, jupyterhub, keycloak

_DISTRIBUTION_METADATA = importlib.metadata.metadata("adalib-auth")
__project__ = _DISTRIBUTION_METADATA["name"]
__version__ = _DISTRIBUTION_METADATA["version"]
__description__ = _DISTRIBUTION_METADATA["description"]

__all__ = ["config", "jupyterhub", "keycloak"]



"""Module to work with JupyterHub on the AdaLab platform.
"""

from .jupyterhub import get_token_from_auth_state, is_this_jupyterhub

__all__ = ["get_token_from_auth_state", "is_this_jupyterhub"]

__title__ = "adalib-auth JupyterHub"



import logging
from typing import Union

import requests

import keycloak
from keycloak.exceptions import KeycloakGetError

from . import config, jupyterhub


def app_authentication(audience: str, scope: str = "openid") -> dict:
    """
    Returns an audience authentication token from an exchange with an app token.

    :param audience: The audience for which the token is requested.
    :type audience: str
    :param scope: The scope of the token exchange, defaults to "openid"
    :type scope: str, optional
    :return: The audience token.
    :rtype: dict
    """
    adalib_config = config.get_config()
    app_token = {
        "access_token": adalib_config.CREDENTIALS["app_access_token"],
        "refresh_token": adalib_config.CREDENTIALS["app_refresh_token"],
    }

    # Exchange the app token for a Jupyterhub token
    jh_token = get_token_from_exchange(
        audience=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
        token=app_token,
        client_id=adalib_config.CREDENTIALS["client_id"],
        client_secret=adalib_config.CREDENTIALS["client_secret"],
        scope=scope,
    )
    update_tokens_in_config(
        new_token=jh_token,
        client_id=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
    )
    # Exchange the Jupyterhub token for an audience token
    audience_token = get_token_from_exchange(
        audience=audience,
        token=jh_token,
        client_id=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
        client_secret=adalib_config.KEYCLOAK_CLIENTS["jupyterhub_secret"],
        scope=scope,
    )
    return audience_token


def authenticate_with_stored_credentials(
    audience_client_id: str,
) -> Union[dict, None]:
    """Attempt to get token for audience_client using stored credentials.

    :param audience_client_id: The client ID of the audience.
    :type audience_client_id: str
    :return: The token for the current user to authenticate to the audience_client with.
    :rtype: dict or None
    """

    adalib_config = config.get_config()
    token = {
        "access_token": adalib_config.CREDENTIALS["access_token"],
        "refresh_token": adalib_config.CREDENTIALS["refresh_token"],
    }
    if token["access_token"] is None and token["refresh_token"] is None:
        logging.debug("No stored credentials.")
        return None
    try:
        audience_token = get_token_from_exchange(
            audience=audience_client_id,
            token=token,
            client_id=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
            client_secret=adalib_config.KEYCLOAK_CLIENTS["jupyterhub_secret"],
        )
    except AssertionError:
        logging.warning("Stored credentials are no longer valid.")
        return None
    return audience_token


def external_authentication(audience: str, scope: str = "openid") -> dict:
    """
    Returns an audience authentication token from an exchange with external user credentials.

    :param audience: The audience for which the token is requested.
    :type audience: str
    :param scope: The scope of the token exchange, defaults to "openid"
    :type scope: str, optional
    :return: The audience token.
    :rtype: dict
    """
    adalib_config = config.get_config()
    resp = requests.get(
        adalib_config.ADABOARD_API_URL + "/adalib/token",
        headers={
            "Authorization": f"Token {adalib_config.CREDENTIALS['token']}"
        },
    )
    if resp.status_code != 200:
        raise AssertionError(
            "Failed to authenticate with external credentials."
        )

    jh_token = resp.json()["access_token"]
    client = get_client(
        client_id=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
        client_secret=adalib_config.KEYCLOAK_CLIENTS["jupyterhub_secret"],
    )
    new_token = client.exchange_token(
        token=jh_token,
        audience=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
    )
    update_tokens_in_config(
        new_token=new_token,
        client_id=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
    )
    audience_token = get_token_from_exchange(
        audience=audience,
        token=new_token,
        client_id=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
        client_secret=adalib_config.KEYCLOAK_CLIENTS["jupyterhub_secret"],
        scope=scope,
    )
    return audience_token


def get_client(
    client_id: str = None,
    client_secret: str = None,
) -> keycloak.KeycloakOpenID:
    """Returns a Keycloak client configured for a specific client. Defaults to
    the Jupyterhub client.

    :param client_id: The ID of a client that is registered in Keycloak
    :param client_secret: The secret matching the given client ID
    :return: Keycloak client for the given client ID.
    :rtype: KeycloakClient
    """
    adalib_config = config.get_config()
    external = True if adalib_config.ENVIRONMENT == "external" else False
    if external:
        server_url = adalib_config.SERVICES["keycloak"]["external"]
    else:
        server_url = adalib_config.SERVICES["keycloak"]["url"]

    try:
        client = keycloak.KeycloakOpenID(
            server_url=server_url,
            client_id=client_id,
            client_secret_key=client_secret,
            realm_name=adalib_config.KEYCLOAK_REALM,
            verify=True,
        )
        client.well_known()  # triggers a request to validate base URL
        return client
    except KeycloakGetError:
        logging.warning(
            "Keycloak client init failed without /auth/, retrying with suffix."
        )
        server_url_with_auth = server_url.rstrip("/") + "/auth/"
        client = keycloak.KeycloakOpenID(
            server_url=server_url_with_auth,
            client_id=client_id,
            client_secret_key=client_secret,
            realm_name=adalib_config.KEYCLOAK_REALM,
            verify=True,
        )
        client.well_known()  # raise if the fallback also fails
        return client


def get_client_token(audience_client_id: str) -> dict:
    """Get the token for the current user to authenticate to adaboard with depending on adalib's
    environment configuration.

    :param audience_client_id: The client ID of the audience.
    :type audience_client_id: str
    :return: The token for the current user to authenticate to adaboard with.
    :rtype: dict
    """

    # First try to authenticate with stored credentials
    logging.debug("Attempting to authenticate with stored credentials.")
    token_from_stored_credentials = authenticate_with_stored_credentials(
        audience_client_id=audience_client_id
    )
    if token_from_stored_credentials is not None:
        return token_from_stored_credentials

    # If that fails, try to authenticate with user credentials
    adalib_config = config.get_config()
    logging.debug("Attempting to authenticate with provided credentials.")
    if adalib_config.ENVIRONMENT == "jupyterhub":
        return jupyterhub_authentication(audience=audience_client_id)
    elif adalib_config.ENVIRONMENT == "nonpub-user-app":
        return app_authentication(audience=audience_client_id)
    elif adalib_config.ENVIRONMENT == "external":
        return external_authentication(audience=audience_client_id)


def get_token_from_exchange(
    audience: str,
    token: str,
    client_id: str,
    client_secret: str,
    scope: str = "openid",
) -> dict:
    """Get a token for the audience given a valid client token.

    Note for developers: The scope defaults to `openid` to be consistent with the
    subsequent call to `keycloakOpenID.exchange_token(...)`. If this default is
    changed it should be done so with respect to this underlying method.

    :param audience: Client that we need to get a token for.
    :type audience: str
    :param token: Token to exchange for audience token.
    :type token: str
    :param client_id: Client ID of the client that the token is for.
    :type client_id: str
    :param client_secret: Client secret of the client that the token is for.
    :type client_secret: str
    :param scope: Scope of token requested, defaults to ""
    :type scope: str, optional
    :return: Token to access audience
    :rtype: dict
    """
    client = get_client(client_id=client_id, client_secret=client_secret)
    if not client.introspect(token=token["access_token"])["active"]:
        assert (
            "refresh_token" in token and token["refresh_token"] is not None
        ), "Access token is no longer valid and a refresh token was not provided"
        try:
            token = client.refresh_token(refresh_token=token["refresh_token"])
        except Exception as exc:
            raise AssertionError("Refresh token has expired.") from exc
        update_tokens_in_config(new_token=token, client_id=client_id)

    audience_token = client.exchange_token(
        token=token["access_token"],
        audience=audience,
        subject=None,
        scope=scope,
    )
    return audience_token


def jupyterhub_authentication(audience: str, scope: str = "openid") -> dict:
    """
    Returns an audience authentication token from an exchange with a Jupyterhub token.

    :param audience: The audience for which the token is requested.
    :type audience: str
    :param scope: The scope of the token exchange, defaults to "openid"
    :type scope: str, optional
    :return: The audience token.
    :rtype: dict
    """
    adalib_config = config.get_config()
    jh_token = jupyterhub.get_token_from_auth_state(
        username=adalib_config.CREDENTIALS["username"],
        jh_token=adalib_config.CREDENTIALS["jh_token"],
        jh_api_url=adalib_config.SERVICES["jupyterhub"]["url"] + "/hub/api",
    )
    update_tokens_in_config(
        new_token=jh_token,
        client_id=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
    )
    audience_token = get_token_from_exchange(
        audience=audience,
        token=jh_token,
        client_id=adalib_config.KEYCLOAK_CLIENTS["jupyterhub"],
        client_secret=adalib_config.KEYCLOAK_CLIENTS["jupyterhub_secret"],
        scope=scope,
    )
    return audience_token


def update_tokens_in_config(new_token: dict, client_id: str):
    """Update the tokens in the config object.

    :param new_token: The new token to update the config with.
    :type new_token: dict
    :param client_id: The client ID of the client that the token is for.
    :type client_id: str
    """
    adalib_config = config.get_config()
    if client_id == adalib_config.KEYCLOAK_CLIENTS["jupyterhub"]:
        adalib_config.CREDENTIALS["access_token"] = new_token["access_token"]
        adalib_config.CREDENTIALS["refresh_token"] = new_token["refresh_token"]



import os

import requests_oauthlib


def get_token_from_auth_state(
    username: str,
    jh_token: str,
    jh_api_url: str,
) -> dict:
    """Get a keycloak token from jupyter auth state.

    :return: Token to access jupyterhub.
    :rtype: dict
    """
    if jh_api_url.startswith("http://"):
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"

    session = requests_oauthlib.OAuth2Session(token={"access_token": jh_token})
    response = session.get(f"{jh_api_url}/users/{username}")
    response.raise_for_status()
    user_info = response.json()

    auth_state = user_info["auth_state"]
    if not auth_state:
        raise ValueError("No auth state returned by JupyterHub API")

    if (
        "JUPYTERHUB_SERVER_NAME" in os.environ
        and os.environ["JUPYTERHUB_SERVER_NAME"] in auth_state
    ):
        return auth_state[os.environ["JUPYTERHUB_SERVER_NAME"]]

    if "" in auth_state:
        return auth_state[""]

    return auth_state


def is_this_jupyterhub():
    """Checks whether this package is running in a JupyterHub environment.

    :return: `true` if called from a program in JupyterHub, `false` otherwise
    :rtype: bool
    """
    return "JUPYTERHUB_API_TOKEN" in os.environ



import logging
import os
import threading
import urllib.parse

import requests


class Singleton(type):
    """Class to ensure singleton behaviour of the Configuration class."""

    _instances = dict()
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        """Instantiate Configuration singleton."""
        t = threading.current_thread()
        if t.native_id not in cls._instances:
            with cls._lock:
                cls._instances[t.native_id] = {
                    "cls": super(Singleton, cls).__call__(*args, **kwargs),
                    "thread": t,
                }

        entries_to_delete = list()
        for k, v in cls._instances.items():
            if not v["thread"].is_alive():
                entries_to_delete.append(k)

        for _key in entries_to_delete:
            del cls._instances[_key]

        return cls._instances[t.native_id]["cls"]

    def reset_instances(cls):
        """Reset Configuration singleton instance."""
        cls._instances = dict()


class Configuration(metaclass=Singleton):
    """Configuration singleton class."""

    def __init__(self, adaboard_api_url=None, **kwargs):
        """Initialize Configuration instance"""
        self.LOG_LEVEL = os.getenv("ADALIB_LOG_LEVEL", "INFO")
        self.HARBOR_AUTH_FILE = os.path.join(
            os.getenv("XDG_RUNTIME_DIR", ".docker"), "config.json"
        )
        self.__configure_logging()
        self.__configure_environment()
        self.__configure_adaboard(adaboard_api_url=adaboard_api_url)
        self.__configure_clients()
        self.__configure_services(**kwargs)
        self.__configure_credentials(**kwargs)
        match self.ENVIRONMENT:
            case "jupyterhub":
                logging.info("adalib configured, running in JupyterHub.")
            case "nonpub-user-app":
                logging.info(
                    "adalib configured, running in a non-public deployed app."
                )
            case "external":
                logging.info(
                    "adalib configured, running in an external environment."
                )

    def __configure_credentials(self, **kwargs):
        """access token and refresh token refer to the JH client"""
        if self.ENVIRONMENT == "jupyterhub":
            self.CREDENTIALS = {
                "username": os.environ["LOGNAME"],
                "jh_token": os.environ["JUPYTERHUB_API_TOKEN"],
                "access_token": None,
                "refresh_token": None,
            }
        elif self.ENVIRONMENT == "nonpub-user-app":
            self.CREDENTIALS = {
                "client_id": os.environ["_UA_CLIENT_ID"],
                "client_secret": os.environ["_UA_CLIENT_SECRET"],
                "app_access_token": kwargs.get(
                    "app_access_token"
                ),  # Deployed app client
                "app_refresh_token": kwargs.get(
                    "app_refresh_token"
                ),  # Deployed app client
                "access_token": None,
                "refresh_token": None,
            }
        elif self.ENVIRONMENT == "external":
            self.CREDENTIALS = {
                "access_token": None,
                "refresh_token": None,
                "token": kwargs.get("token"),
            }

    def __configure_environment(self, **kwargs):
        """
        Find out in which environment is adalib running.
        """
        if "JUPYTERHUB_API_TOKEN" in os.environ:
            self.ENVIRONMENT = "jupyterhub"
        elif "_UA_CLIENT_ID" in os.environ:
            self.ENVIRONMENT = "nonpub-user-app"
        else:
            self.ENVIRONMENT = "external"

    def __configure_adaboard(self, adaboard_api_url=None):
        """Confgiure the target Adaboard API URL."""
        if self.ENVIRONMENT == "external" and (
            adaboard_api_url is None and "ADABOARD_API_URL" not in os.environ
        ):
            raise ValueError(
                "adaboard_api_url must be provided for external environment"
            )
        if adaboard_api_url:
            self.ADABOARD_API_URL = adaboard_api_url
        elif "ADABOARD_API_URL" in os.environ:
            self.ADABOARD_API_URL = os.environ["ADABOARD_API_URL"]
        elif "_NAMESPACE" in os.environ:
            self.ADABOARD_API_URL = (
                f"http://adaboard-api-svc.{os.environ['_NAMESPACE']}:8000"
            )
        else:
            self.ADABOARD_API_URL = "http://adaboard-api-svc:8000"

    def __configure_service_x(self, x: str, given_name: str, apps: list[dict]):
        """Configure AdaLab app or service."""
        url_type = "external" if self.ENVIRONMENT == "external" else "internal"
        for app in apps:
            if app["name"] == x:
                url_parsed = urllib.parse.urlparse(url=app[url_type])
                self.SERVICES[given_name] = {
                    "url": app["internal"],
                    "external": app["external"],
                    "netloc": url_parsed.netloc,
                    "path": url_parsed.path,
                }

    def __configure_services(self, **kwargs):
        """Configure all services present in the AdaLab deployment."""
        token = kwargs.get("token")
        headers = (
            {"Authorization": f"token {token}"} if token is not None else None
        )
        resp = requests.get(
            self.ADABOARD_API_URL,
            params={"include_apps": True},
            timeout=10,
            headers=headers,
        ).json()

        apps = resp["apps"]

        if (
            resp["jh_secret"] == ""
            and "ADALAB_CLIENT_SECRET" not in os.environ
        ):
            raise ValueError("AdaLab client secret was not set")
        elif resp["jh_secret"] == "":
            self.KEYCLOAK_CLIENTS["jupyterhub_secret"] = os.environ[
                "ADALAB_CLIENT_SECRET"
            ]
        else:
            self.KEYCLOAK_CLIENTS["jupyterhub_secret"] = resp["jh_secret"]

        namespace = resp["namespace"]
        network_host = resp["network_host"]

        self.NAMESPACE = namespace
        self.NETWORK_HOST = network_host
        self.SERVICES = dict()
        self.__configure_service_x(
            x="adaboard-api", given_name="adaboard-api", apps=apps
        )
        self.__configure_service_x(
            x="container-registry", given_name="harbor", apps=apps
        )
        self.__configure_service_x(
            x="jupyterhub", given_name="jupyterhub", apps=apps
        )
        self.__configure_service_x(
            x="keycloak", given_name="keycloak", apps=apps
        )
        self.__configure_service_x(x="mlflow", given_name="mlflow", apps=apps)
        self.__configure_service_x(
            x="superset", given_name="superset", apps=apps
        )

    def __configure_clients(self):
        """Configure clients for services present in the AdaLab deployment."""
        self.KEYCLOAK_REALM = os.environ.get("KEYCLOAK_REALM", "adalab")
        self.KEYCLOAK_CLIENTS = dict()
        self.KEYCLOAK_CLIENTS["adaboard-api"] = os.environ.get(
            "KEYCLOAK_ADABOARD_CLIENT_ID", "adaboard"
        )
        self.KEYCLOAK_CLIENTS["harbor"] = os.environ.get(
            "KEYCLOAK_HARBOR_CLIENT_ID", "harbor"
        )
        self.KEYCLOAK_CLIENTS["jupyterhub"] = os.environ.get(
            "KEYCLOAK_JUPYTERHUB_CLIENT_ID", "jupyterhub"
        )
        self.KEYCLOAK_CLIENTS["superset"] = os.environ.get(
            "KEYCLOAK_SUPERSET_CLIENT_ID", "superset"
        )

    def __configure_logging(self):
        """Set up execution logs."""
        log_level = getattr(logging, self.LOG_LEVEL.upper(), logging.INFO)
        logging.basicConfig(
            level=log_level,
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        )

    @classmethod
    def clean(cls):
        """
        Delete existing instances of the class.
        """
        cls.reset_instances()

    def reset(self, adaboard_api_url=None, **kwargs):
        """
        Clean up and create a new class instance with the new parameters.
        """
        self.clean()
        self.__init__(adaboard_api_url=adaboard_api_url, **kwargs)


def get_config(adaboard_api_url=None, **kwargs):
    """
    Initialize the Configuration singleton. This is called automatically when
    the Configuration class is imported.
    """
    return Configuration(adaboard_api_url=adaboard_api_url, **kwargs)
